"""
Miscelenious user profile commands
"""
from firecloud import fiss


def list_workspace_namespaces() -> list:
    """
    Billing projects available to the current usuer
    """
    resp = fiss.fapi.list_billing_projects()
    resp.raise_for_status()
    return resp.json()



import json
import argparse
from uuid import uuid4

from terra_notebook_utils import table as tnu_table
from terra_notebook_utils.cli import dispatch, Config


table_cli = dispatch.group("table", help=tnu_table.__doc__, arguments={
    "--workspace": dict(
        type=str,
        default=None,
        help="workspace name. If not provided, the configured CLI workspace will be used"
    ),
    "--workspace-namespace": dict(
        type=str,
        default=None,
        help=("workspace namespace. If not provided, the configured CLI google billing "
              "project will be used.")
    ),
})

@table_cli.command("list")
def list_tables(args: argparse.Namespace):
    """
    List all tables in the workspace
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    for table in tnu_table.list_tables(args.workspace, args.workspace_namespace):
        print(table)

@table_cli.command("list-rows", arguments={
    "--table": dict(type=str, required=True, help="table name")
})
def list_rows(args: argparse.Namespace):
    """
    Get all rows
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    for row in tnu_table.list_rows(args.table, args.workspace, args.workspace_namespace):
        print(row.name, row.attributes)

@table_cli.command("get-row", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--row": dict(type=str, required=True, help="row name"),
})
def get_row(args: argparse.Namespace):
    """
    Get one row
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    row = tnu_table.get_row(args.table, args.row, args.workspace, args.workspace_namespace)
    if row is not None:
        print(row.name, json.dumps(row.attributes))

@table_cli.command("delete-table", arguments={
    "--table": dict(type=str, required=True, help="table name"),
})
def delete_table(args: argparse.Namespace):
    """
    Get one row
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    tnu_table.delete(args.table, args.workspace, args.workspace_namespace)

@table_cli.command("delete-row", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--row": dict(type=str, required=True, help="row name"),
})
def delete_row(args: argparse.Namespace):
    """
    Delete a row
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    tnu_table.del_row(args.table, args.row, args.workspace, args.workspace_namespace)

@table_cli.command("fetch-drs-url", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--file-name": dict(type=str, required=True, help="file name"),
})
def fetch_drs_url(args: argparse.Namespace):
    """
    Fetch the DRS URL associated with `--file-name` in `--table`.
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    print(tnu_table.fetch_drs_url(args.table, args.file_name, args.workspace, args.workspace_namespace))

@table_cli.command("put-row", arguments={
    "--table": dict(type=str, required=True, help="table name"),
    "--row": dict(type=str, required=False, default=None, help="row name. This should be unique for the table"),
    "data": dict(type=str, nargs="+", help="list of key-value pairs")
})
def put_row(args: argparse.Namespace):
    """
    Put a row.
    Example:
    tnu table put-row \\
    --table abbrv_merge_input \\
    --id 1 \\
    bucket=fc-9169fcd1-92ce-4d60-9d2d-d19fd326ff10 \\
    input_keys=test_vcfs/a.vcf.gz,test_vcfs/b.vcf.gz \\
    output_key=foo.vcf.gz
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    attributes = dict()
    for pair in args.data:
        key, val = pair.split("=")
        attributes[key] = val
    row = tnu_table.Row(name=args.row or f"{uuid4()}", attributes=attributes)
    tnu_table.put_row(args.table, row, args.workspace, args.workspace_namespace)



import os
import gzip
import typing
import tarfile
from math import ceil
from pathlib import Path
from contextlib import closing
from typing import Optional

import gs_chunked_io as gscio
from google.cloud.storage.bucket import Bucket

from terra_notebook_utils import gs
from terra_notebook_utils.progress import ProgressBar

_chunk_size = 1024 * 1024 * 32

# Set this for unit tests only
_extract_single_chunk: bool = False

def extract(src_fh, dst_bucket: typing.Optional[Bucket]=None, root: typing.Optional[str]=None):
    """
    Extract a tar.gz archive into the local filesystem, or a GS bucket if `dst_bucket` is provided.
    """
    gzip_reader = gzip.GzipFile(fileobj=src_fh)
    tf = tarfile.TarFile(fileobj=gzip_reader)  # type: ignore
    for tarinfo in tf:
        if tarinfo.isfile():
            if dst_bucket:
                dst_fh = _prepare_gs(tarinfo, dst_bucket, root or "")
            else:
                dst_fh = _prepare_local(tarinfo, root or "")
            print(f"Inflating {tarinfo.name}")
            progress_bar = ProgressBar(ceil(tarinfo.size / _chunk_size) + 1,
                                       size=ceil(tarinfo.size / 1024 ** 2),
                                       units="MB")
            with closing(progress_bar):
                with closing(dst_fh):
                    _transfer_data(tf.extractfile(tarinfo), dst_fh, progress_bar)
                progress_bar.update()

            # Hack for speedy unit tests
            if _extract_single_chunk:
                return

def _transfer_data(from_fh, to_fh, progress_bar: ProgressBar=None):
    while True:
        data = from_fh.read(_chunk_size)
        if data:
            to_fh.write(data)
        else:
            break
        if progress_bar:
            progress_bar.update()

        # Hack for speedy unit tests
        if _extract_single_chunk:
            return

def _prepare_local(tarinfo: tarfile.TarInfo, root: str):
    filepath = os.path.abspath(os.path.join(root, tarinfo.name))
    Path(os.path.dirname(filepath)).mkdir(parents=True, exist_ok=True)
    return open(filepath, "wb")

def _prepare_gs(tarinfo: tarfile.TarInfo, bucket: Bucket, root: typing.Optional[str]):
    if root is not None:
        key = f"{root}/{tarinfo.name}"
    else:
        key = tarinfo.name
    return gscio.Writer(key, bucket, chunk_size=_chunk_size)



import sys
import json
import argparse
from typing import Any, Dict

from terra_notebook_utils import drs, MULTIPART_THRESHOLD
from terra_notebook_utils.cli import dispatch, Config
from terra_notebook_utils.drs import DRSResolutionError


drs_cli = dispatch.group("drs", help=drs.__doc__)

workspace_args: Dict[str, Dict[str, Any]] = {
    "--workspace": dict(
        type=str,
        default=None,
        help="workspace name. If not provided, the configured CLI workspace will be used"
    ),
    "--workspace-namespace": dict(
        type=str,
        required=False,
        default=Config.info['workspace_namespace'],
        help=("The billing project for GS requests. "
              "If omitted, the CLI configured `workspace_namespace` will be used. "
              "Note that DRS URLs also involve a GS request.")
    )
}

@drs_cli.command("copy", arguments={
    "drs_url": dict(type=str),
    "dst": dict(type=str, help="local file path, or Google Storage location if prefixed with 'gs://'"),
    ** workspace_args,
})
def drs_copy(args: argparse.Namespace):
    """
    Copy drs:// object to local file or Google Storage bucket
    examples:
        tnu drs copy drs://my-drs-id /tmp/doom
        tnu drs copy drs://my-drs-id gs://my-cool-bucket/my-cool-bucket-key
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    drs.copy(args.drs_url, args.dst, args.workspace, args.workspace_namespace)

@drs_cli.command("copy-batch", arguments={
    "drs_urls": dict(type=str, nargs="*", help="space separated list of drs:// URIs"),
    "--dst": dict(type=str, required=True, help="local directory, or Google Storage location if prefixed with 'gs://'"),
    ** workspace_args,
})
def drs_copy_batch(args: argparse.Namespace):
    """
    Copy several drs:// objects to local directory or Google Storage bucket
    examples:
        tnu drs copy drs://my-drs-1 drs://my-drs-2 drs://my-drs-3 --dst /tmp/doom
        tnu drs copy drs://my-drs-1 drs://my-drs-2 drs://my-drs-3 --dst gs://my-cool-bucket/my-cool-folder
    """
    assert 1 <= len(args.drs_urls)
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    drs.copy_batch(args.drs_urls, args.dst, args.workspace, args.workspace_namespace)

@drs_cli.command("head", arguments={
    "drs_url": dict(type=str),
    "--bytes": dict(type=int, required=False, default=1, help="Number of bytes to fetch."),
    "--buffer": dict(type=int, required=False, default=MULTIPART_THRESHOLD,
                     help="Control the buffer size when fetching data."),
    ** workspace_args,
})
def drs_head(args: argparse.Namespace):
    """
    Print the first bytes of a drs:// object.

    Example:
        tnu drs head drs://crouching-drs-hidden-access
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    the_bytes = drs.head(args.drs_url,
                         num_bytes=args.bytes,
                         buffer=args.buffer,
                         workspace_name=args.workspace,
                         workspace_namespace=args.workspace_namespace)
    # sys.stdout.buffer is used outside of a python notebook since that's the standard non-lossy way
    # to write/display bytes; sys.stdout.buffer is not available inside of a python notebook
    # though, as sys.stdout is a ipykernel.iostream.OutStream object:
    # https://github.com/ipython/ipykernel/blob/master/ipykernel/iostream.py#L265
    # so we use bare sys.stdout and rely on the ipykernel method's lossy unicode stream
    stdout_buffer = getattr(sys.stdout, 'buffer', sys.stdout)
    stdout_buffer.write(the_bytes)

@drs_cli.command("extract-tar-gz", arguments={
    "drs_url": dict(type=str),
    "dst_gs_url": dict(type=str, help=("Root of extracted archive. This must be a Google Storage location prefixed"
                                       "prefixed with 'gs://'")),
    ** workspace_args,
})
def drs_extract_tar_gz(args: argparse.Namespace):
    """
    Extract a `tar.gz` archive resolved by DRS into a Google Storage bucket.
    example:
        tnu drs extract-tar-gz drs://my-tar-gz gs://my-dst-bucket/root
    """
    assert args.dst_gs_url.startswith("gs://")
    bucket, pfx = args.dst_gs_url[5:].split("/", 1)
    pfx = pfx or None
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    drs.extract_tar_gz(args.drs_url, pfx, bucket, args.workspace, args.workspace_namespace)

@drs_cli.command("info", arguments={
    "drs_url": dict(type=str),
})
def drs_info(args: argparse.Namespace):
    """
    Get information about drs:// objects
    """
    info = drs.info(args.drs_url)
    print(json.dumps(info, indent=2))

@drs_cli.command("credentials", arguments={
    "drs_url": dict(type=str),
})
def drs_credentials(args: argparse.Namespace):
    """
    Return the credentials needed to access a DRS url.
    """
    try:
        info = drs.resolve_drs_info_for_gs_storage(args.drs_url)
    except DRSResolutionError:
        raise
    except Exception:
        raise

    print(json.dumps(info.credentials, indent=2))



from math import ceil


SECOND = 1.0
MINUTE = SECOND * 60
HOUR = MINUTE * 60
DAY = HOUR * 24
MONTH = DAY * 30.437  # Not critical to be exact

class PersistentDisk:
    """
    GCP persistent disk costs:
    https://cloud.google.com/compute/disks-image-pricing#persistentdisk
    """
    standard = 0.04 / MONTH  # GCP disk prices are quoted per month :(

    @classmethod
    def estimate(cls, size_gb: float, runtime_seconds: float) -> float:
        return cls.standard * size_gb * runtime_seconds

class GCPCustomN1Cost:
    # Pricing information: https://cloud.google.com/compute/vm-instance-pricing#n1_custommachinetypepricing
    on_demand_per_cpu = 0.033174 / HOUR
    on_demand_per_gb = 0.004446 / HOUR
    preemptible_per_cpu = 0.00698 / HOUR
    preemptible_per_gb = 0.00094 / HOUR

    @classmethod
    def estimate(cls, cpus: int, memory_gb: float, runtime_seconds: float, preemptible: bool) -> float:
        # GCP Instance Billing Model:
        # https://cloud.google.com/compute/vm-instance-pricing#billingmodel
        if preemptible:
            per_cpu = cls.preemptible_per_cpu
            per_gb = cls.preemptible_per_gb
        else:
            per_cpu = cls.on_demand_per_cpu
            per_gb = cls.on_demand_per_gb
        cost_duration_hours = ceil(MINUTE if MINUTE >= runtime_seconds else runtime_seconds)
        cost_for_cpus = cost_duration_hours * cpus * per_cpu
        cost_for_mem = cost_duration_hours * memory_gb * per_gb
        cost = cost_for_cpus + cost_for_mem
        return cost



import json
import argparse

from terra_notebook_utils import profile
from terra_notebook_utils.cli import dispatch


profile_cli = dispatch.group("profile", help=profile.__doc__)


@profile_cli.command("list-workspace-namespaces")
def list_workspace_namespaces(args: argparse.Namespace):
    """
    Workspace namespaces available to the current usuer
    """
    print(json.dumps(profile.list_workspace_namespaces(), indent=2))



"""
Terra data table commands
"""
import os
from uuid import uuid4
from collections import defaultdict, namedtuple
from typing import Any, Dict, Generator, Iterable, List, Optional, Set, Tuple, Union

import requests
from firecloud import fiss

from terra_notebook_utils import WORKSPACE_GOOGLE_PROJECT, WORKSPACE_NAME
from terra_notebook_utils.utils import _AsyncContextManager


Row = namedtuple("Row", "name attributes")

VALUE = Union[str, int, float, bool]
ATTRIBUTES = Dict[str, Union[VALUE, Iterable[VALUE]]]
UPDATE_OPS = List[Dict[str, Any]]
COLUMN_HEADERS = Tuple[str, ...]
ROW_LIKE = Tuple[str, ATTRIBUTES]
ROW_OR_NAME = Union[Row, str]

# A note on types:
# As presented in Terra's UI, tables may contain string, integer, and boolean values.  However, Firecloud API clients
# must transmit tables using TSV formatted data, which is not typed. So, alas, we cannot transmit typed data to Terra
# data tables, and cannot transform uploaded types via the Firecloud API (as far as I know).
#
# It would be preferable if the Firecloud API exposed JSON endpoints for uploading table data (which may already exist
# on the backend?)

# It turns out google.auth.transport.requests.AuthorizedSession is not thread safe.
# Fortunately fiss.fapi._set_session caches the result. Call it once on the main thread.
fiss.fapi._set_session()

class Writer(_AsyncContextManager):
    """
    Distribute row uploads across as few API calls as possible.
    Uploads are performed in the background.
    Also transparently handles sequences, which the Firecloud API makes difficult. (These must be uploaded and modified
    via separate API calls.)
    """
    def __init__(self,
                 name: str,
                 workspace: Optional[str]=WORKSPACE_NAME,
                 workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT,
                 tsv_upload_size: int=1024):
        self.name = name
        self._init_request_data()
        self._workspace = workspace
        self._workspace_google_project = workspace_namespace
        self._tsv_upload_size = tsv_upload_size

    def _init_request_data(self):
        self._tsvs: Dict[COLUMN_HEADERS, str] = defaultdict(list)
        self._row_update_request_data: Dict[COLUMN_HEADERS, List[Tuple[Row, UPDATE_OPS]]] = dict()

    def _get_row_update_request_data(self, row: Row) -> List[Dict[str, Any]]:
        request_data = list()
        for name, val in row.attributes.items():
            if isinstance(val, str):
                update_ops = list()  # No Firecloud update operations needed for string values
            elif isinstance(val, (int, float, bool)):
                update_ops = [dict(op="AddUpdateAttribute", attributeName=name, addUpdateAttribute=val)]
            elif hasattr(val, "__iter__"):
                update_ops = [dict(op="RemoveAttribute", attributeName=name)]
                types: Set[type] = set()
                for m in val:
                    update_ops.append(dict(op="AddListMember", attributeListName=name, newMember=m))
                    types.add(type(m))
                assert 1 == len(types)
            request_data.extend(update_ops)
        return request_data

    def put_row(self, item: Union[ROW_LIKE, ATTRIBUTES]) -> str:
        if isinstance(item, dict):
            row = Row(f"{uuid4()}", item)
        else:
            row = Row(*item)
        column_headers = tuple(sorted(row.attributes.keys()))
        if column_headers not in self._tsvs:
            self._tsvs[column_headers] = "\t".join([f"{self.name}_id", *column_headers])
            self._row_update_request_data[column_headers] = list()
        self._tsvs[column_headers] += (
            os.linesep
            + row.name + "\t"
            + "\t".join(row.attributes[c] if isinstance(row.attributes[c], str)
                        else "x"  # Dummy value to be replaced during row update API calls.
                        for c in column_headers)
        )
        update_request_data = self._get_row_update_request_data(row)
        if update_request_data:
            self._row_update_request_data[column_headers].append((row, update_request_data))
        self._upload()
        return row.name

    def _upload(self, force: bool=False):
        """
        Schedule uploads for all TSVs of size equal to or greater than `self._tsv_upload_size`.
        If `force` is True, schedule uploads for all TSVs.
        """
        for column_headers, tsv in self._tsvs.copy().items():
            if force or len(tsv) >= self._tsv_upload_size:
                row_updates = self._row_update_request_data.get(column_headers, list())
                self.submit(self._do_fiss_upload, tsv, row_updates)
                del self._tsvs[column_headers]
                if row_updates:
                    del self._row_update_request_data[column_headers]

    def _do_fiss_upload(self, tsv: str, row_update_request_data: List[Tuple[Row, List[Dict[str, Any]]]]):
        fiss.fapi.upload_entities(self._workspace_google_project,
                                  self._workspace,
                                  tsv,
                                  model="flexible").raise_for_status()
        for row, request_data in row_update_request_data:
            if request_data:
                self.submit(self._do_fiss_updates, row, request_data)

    def _do_fiss_updates(self, row: Row, request_data: UPDATE_OPS, retry: int=0):
        try:
            fiss.fapi.update_entity(self._workspace_google_project,
                                    self._workspace,
                                    self.name,
                                    row.name,
                                    request_data).raise_for_status()
        except requests.exceptions.HTTPError as e:
            if 500 == e.response.status_code:
                # Firecloud occasionally throws 500 errors for successful update operations.
                # Check if we get the row we expect, retry otherwise.
                if 5 > retry:
                    cur_row = get_row(self.name, row.name)
                    if row != cur_row:
                        self.submit(self._do_fiss_updates, row, request_data, retry=retry + 1)
                else:
                    raise Exception(f"Ran out of retries updating row {row.name}") from e
            else:
                raise

    def _prepare_for_exit(self):
        if self._tsvs:
            self._upload(force=True)

class Deleter(_AsyncContextManager):
    """
    Distribute row deletes across as few API calls as possible.
    """
    def __init__(self,
                 name: str,
                 workspace: Optional[str]=WORKSPACE_NAME,
                 workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT,
                 rows_per_api_call: int=2000):
        self.name = name
        self._workspace = workspace
        self._workspace_google_project = workspace_namespace
        self._init_request_data()
        self._rows_per_api_call = rows_per_api_call

    def _init_request_data(self):
        self._request_data: List[Dict[str, str]] = list()

    def del_row(self, item: Union[str, ROW_LIKE]):
        if isinstance(item, str):
            name = item
        elif isinstance(item, Row):
            name = item.name
        else:
            name = item[0]
        self._request_data.append(dict(entityType=self.name, entityName=name))
        if self._rows_per_api_call <= len(self._request_data):
            self._delete()

    def _delete(self):
        self.submit(self._do_fiss_delete, self._request_data)
        self._init_request_data()

    def _do_fiss_delete(self, ents: List[Dict[str, str]]):
        try:
            fiss.fapi.delete_entities(self._workspace_google_project,
                                      self._workspace,
                                      ents).raise_for_status()
        except requests.exceptions.HTTPError as e:
            if 400 == e.response.status_code:
                pass
            else:
                raise

    def _prepare_for_exit(self):
        if self._request_data:
            self._delete()

def list_tables(workspace: Optional[str]=WORKSPACE_NAME,
                workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Generator[str, None, None]:
    resp = fiss.fapi.list_entity_types(workspace_namespace, workspace)
    resp.raise_for_status()
    for table_name in resp.json():
        yield table_name

def _get_rows_page(table: str,
                   page_number: int=1,
                   page_size: int=500,
                   sort_direction: str="asc",
                   filter_terms: Any=None,
                   workspace: Optional[str]=WORKSPACE_NAME,
                   workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Tuple[int, dict]:
    resp = fiss.fapi.get_entities_query(workspace_namespace,
                                        workspace,
                                        table,
                                        page=page_number,
                                        page_size=page_size,
                                        sort_direction=sort_direction,
                                        filter_terms=filter_terms)
    resp.raise_for_status()
    body = resp.json()
    return body['resultMetadata']['filteredPageCount'], body['results']

def _attributes_from_fiss_response(attributes: Dict[str, Any]) -> ATTRIBUTES:
    return {key: val['items'] if isinstance(val, dict) and "items" in val else val
            for key, val in attributes.items()}

def list_rows(table: str,
              workspace: Optional[str]=WORKSPACE_NAME,
              workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Generator[Row, None, None]:
    page_number = total_pages = 1
    while total_pages >= page_number:
        total_pages, rows = _get_rows_page(table,
                                           page_number,
                                           workspace=workspace,
                                           workspace_namespace=workspace_namespace)
        for item in rows:
            yield Row(item['name'], _attributes_from_fiss_response(item['attributes']))
        page_number += 1

def get_row(table: str,
            item: ROW_OR_NAME,
            workspace: Optional[str]=WORKSPACE_NAME,
            workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Optional[Row]:
    row_name = item.name if isinstance(item, Row) else item
    resp = fiss.fapi.get_entity(workspace_namespace, workspace, table, row_name)
    try:
        resp.raise_for_status()
    except requests.exceptions.HTTPError as e:
        if 404 == e.response.status_code:
            return None
        else:
            raise
    data = resp.json()
    return Row(data['name'], _attributes_from_fiss_response(data['attributes']))

def put_rows(table: str,
             items: Iterable[Union[ROW_LIKE, ATTRIBUTES]],
             workspace: Optional[str]=WORKSPACE_NAME,
             workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> List[str]:
    with Writer(table, workspace, workspace_namespace) as tw:
        return [tw.put_row(item) for item in items]

def put_row(table: str,
            item: Union[ROW_LIKE, ATTRIBUTES],
            workspace: Optional[str]=WORKSPACE_NAME,
            workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> str:
    return put_rows(table, [item], workspace, workspace_namespace)[0]

def del_rows(table: str,
             items: Iterable[ROW_OR_NAME],
             workspace: Optional[str]=WORKSPACE_NAME,
             workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    with Deleter(table, workspace, workspace_namespace) as td:
        for row in items:
            td.del_row(row)

def del_row(table: str,
            item: ROW_OR_NAME,
            workspace: Optional[str]=WORKSPACE_NAME,
            workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    del_rows(table, [item], workspace, workspace_namespace)


def delete(table: str,
           workspace: Optional[str]=WORKSPACE_NAME,
           workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT):
    with Deleter(table, workspace, workspace_namespace) as td:
        for row in list_rows(table, workspace, workspace_namespace):
            td.del_row(row)

def fetch_drs_url(table: str,
                  file_name: str,
                  workspace: Optional[str]=WORKSPACE_NAME,
                  workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> str:
    """
    Fetch the first `object_id` associated with `pfb:file_name` from `table`.
    DRS urls, when available, are stored in `pfb:object_id`.
    Note: prior to 21-October, 2020, column headers omitted the "pfb:" prefix. For the time being, both formats are
          supported.
    """
    for pfx in ("pfb:", ""):
        try:
            for row in list_rows(table, workspace, workspace_namespace):
                if file_name == row.attributes[f"{pfx}file_name"]:
                    val = row.attributes[f"{pfx}object_id"]
                    if isinstance(val, str) and val.startswith("drs://"):
                        return val
                    else:
                        raise ValueError(f"Expected DRS url in '{table}' for '{file_name}'"
                                         f", got '{val}' instead.")
        except KeyError:
            pass
    else:
        raise KeyError(f"Unable to fetch DRS URL for table '{table}', file_name '{file_name}'")



"""
Grab the first N lines of a VCF and gzip or bgzip the result
"""
import io
import os
import sys
import gzip
from contextlib import closing
from functools import lru_cache
from typing import Iterable

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))  # noqa
sys.path.insert(0, pkg_root)  # noqa

import bgzip
from bgzip.bgzip_utils import BGZIPMalformedHeaderException
import gs_chunked_io as gscio

from terra_notebook_utils import gs, drs, WORKSPACE_GOOGLE_PROJECT


def partialize_vcf(uri: str, number_of_lines: int, zip_format: str="bgzip") -> bytes:
    with io.BytesIO() as raw:
        if "bgzip" == zip_format:
            zip_writer = bgzip.BGZipWriter(raw)
        elif "gzip" == zip_format:
            zip_writer = gzip.GzipFile(fileobj=raw, mode="w")
        else:
            raise ValueError("Supported values for `zip_format` are 'bgzip' and 'gzip'")
        with zip_writer as writer:
            for line in _vcf_lines(uri, number_of_lines):
                writer.write(line)
        out = raw.getvalue()
    return out

@lru_cache()
def _vcf_lines(uri: str, number_of_lines: int) -> Iterable[bytes]:
    with closing(_get_fileobj(uri)) as fh:
        if uri.endswith(".gz"):
            with bgzip.BGZipReader(fh) as gzip_reader:
                return [line for _, line in zip(range(number_of_lines), io.BufferedReader(gzip_reader))]
        else:
            return [line for _, line in zip(range(number_of_lines), io.BufferedReader(fh))]

def _get_fileobj(uri: str):
    if uri.startswith("gs://"):
        bucket_name, key = uri[5:].split("/", 1)
        blob = gs.get_client().bucket(bucket_name, user_project=WORKSPACE_GOOGLE_PROJECT).get_blob(key)
        fh = gscio.Reader(blob, chunk_size=1024 ** 2)
    elif uri.startswith("drs://"):
        gs_client, drs_info = drs.resolve_drs_for_gs_storage(uri)
        bucket = gs_client.bucket(drs_info.bucket_name, user_project=WORKSPACE_GOOGLE_PROJECT)
        fh = gscio.Reader(bucket.get_blob(drs_info.key), chunk_size=1024 ** 2)
    else:
        fh = open(uri, "rb")
    return fh



import json
import threading
from concurrent.futures import ThreadPoolExecutor, Future, as_completed
from typing import Any, Callable, Dict, Optional, Iterable, Set

import jmespath


class _AsyncContextManager:
    """
    Context manager for asynchronous execution. Wait on exit for all jobs to complete.
    """
    def submit(self, func: Callable[..., Any], *args, **kwargs):
        f = self._executor.submit(func, *args, **kwargs)
        self._futures.add(f)
        if threading.current_thread() is threading.main_thread():
            self.prune_futures()

    def prune_futures(self):
        assert threading.current_thread() is threading.main_thread()
        for f in self._futures.copy():
            if f.done():
                f.result()
                self._futures.remove(f)

    def _prepare_for_exit(self):
        raise NotImplementedError()

    def __enter__(self):
        self._executor = ThreadPoolExecutor(max_workers=10)
        self._futures: Set[Future] = set()
        return self

    def __exit__(self, *args, **kwargs):
        self._prepare_for_exit()
        while self._futures:
            for f in as_completed(self._futures):
                pass
            self.prune_futures()
        self._executor.shutdown()

def concurrent_recursion(recurse: Callable[[Any], Iterable[Any]], initial_data: Iterable[Any], concurrency: int=8):
    """
    Call `recurse` on each item in `initial_data`, and on each item returned by `recurse`, concurrently.
    """
    with ThreadPoolExecutor(max_workers=concurrency) as e:
        futures = {e.submit(recurse, item) for item in initial_data}
        while futures:
            for f in as_completed(futures):
                futures.remove(f)
                futures.update({e.submit(recurse, item) for item in f.result()})

def js_get(path: str, data: Dict[str, Any], default: Optional[Any]=None) -> Any:
    res = jmespath.search(path, data)
    if res is not None:
        return res
    elif default is not None:
        return default
    else:
        raise KeyError(f"'{path}' not found in {json.dumps(data, indent=2)}")



"""
Configure the CLI
"""
import typing
import argparse

from terra_notebook_utils.cli import Config, dispatch


config_cli = dispatch.group("config", help=__doc__)


@config_cli.command("set-workspace", arguments={
    "workspace": dict()
})
def set_config_workspace(args: argparse.Namespace):
    """
    Set workspace for cli commands
    """
    Config.info["workspace"] = args.workspace
    Config.write()

@config_cli.command("set-workspace-namespace", arguments={
    "workspace_namespace": dict(type=str)
})
def set_config_workspace_namespace(args: argparse.Namespace):
    """
    Set workspace namespace for cli commands
    """
    Config.info["workspace_namespace"] = args.workspace_namespace
    Config.write()

@config_cli.command("print")
def config_print(args: argparse.Namespace):
    """
    Print the CLI configuration.
    """
    import json
    print(json.dumps(Config.info, indent=2))



import io
import warnings

from google.cloud.storage import Client


class SuppressWarningsMixin:
    def setUp(self):
        # Suppress the annoying google gcloud _CLOUD_SDK_CREDENTIALS_WARNING warnings
        warnings.filterwarnings("ignore", "Your application has authenticated using end user credentials")
        # Suppress unclosed socket warnings
        warnings.simplefilter("ignore", ResourceWarning)


def upload_data(uri: str, data: bytes):
    if uri.startswith("gs://"):
        bucket, key = uri[5:].split("/", 1)
        Client().bucket(bucket).blob(key).upload_from_file(io.BytesIO(data))
    else:
        with open(uri, "wb") as fh:
            fh.write(data)



import json
import typing
import argparse

from firecloud import fiss

from terra_notebook_utils import workspace
from terra_notebook_utils.cli import Config, dispatch


workspace_cli = dispatch.group("workspace", help=workspace.__doc__)


@workspace_cli.command("list")
def list_workspaces(args: argparse.Namespace):
    """
    List workspaces available to the current usuer
    """
    list_workspaces.__doc__ = workspace.list_workspaces.__doc__
    data = workspace.list_workspaces()
    info_keys = ["name", "createdBy", "bucketName", "namespace"]
    workspaces = [{key: ws['workspace'][key] for key in info_keys}
                  for ws in data]
    print(json.dumps(workspaces, indent=2))

@workspace_cli.command("get", arguments={
    "--workspace": dict(type=str, required=True, help="workspace name"),
    "--workspace-namespace": dict(type=str, required=False, default=None, help="workspace namespace"),
})
def get_workspace(args: argparse.Namespace):
    """
    Get information about a workspace
    """
    data = workspace.get_workspace(args.workspace, args.workspace_namespace)
    print(json.dumps(data, indent=2))

@workspace_cli.command("get-bucket", arguments={
    "--workspace": dict(type=str, required=False, default=Config.info['workspace'], help="workspace name"),
    "--workspace-namespace": dict(type=str, required=False, default=None, help="workspace namespace"),
})
def get_workspace_bucket(args: argparse.Namespace):
    """
    Get workspace bucket
    """
    data = workspace.get_workspace(args.workspace, args.workspace_namespace)
    bucket_name = data.get('workspace', dict()).get('bucketName', None)
    print(bucket_name)

@workspace_cli.command("delete-workflow-logs", arguments={
    "--workspace": dict(type=str,
                        default=Config.info['workspace'],
                        help="If ommitted, the CLI configured workspace will be used.")
})
def delete_workflow_logs(args: argparse.Namespace):
    """
    THIS CANNOT BE UNDONE
    Delete _ALL_ workflow exeuction logs from the workspace bucket.
    """
    bucket = workspace.get_workspace_bucket(args.workspace)
    workspace.remove_workflow_logs(bucket)



"""
Workflow information
"""
import json
import logging
from datetime import datetime
from functools import lru_cache
from typing import Dict, Generator, Optional, Tuple

from firecloud import fiss

from terra_notebook_utils import WORKSPACE_NAME, WORKSPACE_GOOGLE_PROJECT, costs
from terra_notebook_utils.utils import concurrent_recursion, js_get


logger = logging.getLogger(__name__)

date_format = "%Y-%m-%dT%H:%M:%S.%fZ"

class TNUCostException(Exception):
    pass

def list_submissions(workspace_name: Optional[str]=WORKSPACE_NAME,
                     workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Generator[dict, None, None]:
    resp = fiss.fapi.list_submissions(workspace_namespace, workspace_name)
    resp.raise_for_status()
    for s in resp.json():
        yield s

@lru_cache()
def get_submission(submission_id: str,
                   workspace_name: Optional[str]=WORKSPACE_NAME,
                   workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> dict:
    """
    Get information about a submission, including member workflows
    """
    resp = fiss.fapi.get_submission(workspace_namespace, workspace_name, submission_id)
    resp.raise_for_status()
    return resp.json()

@lru_cache()
def get_workflow(submission_id: str,
                 workflow_id: str,
                 workspace_name: Optional[str]=WORKSPACE_NAME,
                 workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> dict:
    """
    Get information about a workflow
    """
    resp = fiss.fapi.get_workflow_metadata(workspace_namespace, workspace_name, submission_id, workflow_id)
    resp.raise_for_status()
    return resp.json()

def get_all_workflows(submission_id: str,
                      workspace: Optional[str]=WORKSPACE_NAME,
                      workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> Dict[str, dict]:
    """
    Retrieve all workflows, and workflow metadata, for `submission_id`, including sub-workflows.
    """
    workflows_metadata = dict()

    def get_metadata_and_subworkflows(workflow_id: str):
        wf_medadata = get_workflow(submission_id, workflow_id, workspace, workspace_namespace)
        workflows_metadata[workflow_id] = wf_medadata
        subworkflows = {call_metadata['subWorkflowId']
                        for call_metadata_list in wf_medadata['calls'].values()
                        for call_metadata in call_metadata_list
                        if "subWorkflowId" in call_metadata}
        return subworkflows

    submission = get_submission(submission_id, workspace, workspace_namespace)
    initial_workflow_ids = {wf['workflowId'] for wf in submission['workflows']}
    concurrent_recursion(get_metadata_and_subworkflows, initial_workflow_ids)

    return workflows_metadata

def estimate_workflow_cost(workflow_id: str, workflow_metadata: dict) -> Generator[dict, None, None]:
    for call_name, call_metadata_list in workflow_metadata['calls'].items():
        for call_metadata in call_metadata_list:
            if "subWorkflowId" in call_metadata:
                # subworkflows need to be looked up and estimated separately
                continue
            try:
                task_name = call_name.split(".")[1]
                call_cached = bool(int(js_get("callCaching.hit", call_metadata, default=0)))
                if call_cached:
                    cost, cpus, memory_gb, runtime, disk_size_gb = 0.0, 0, 0.0, 0.0, 0.0
                else:
                    cpus, memory_gb = _parse_machine_type(js_get("jes.machineType", call_metadata))
                    # Assume that Google Lifesciences Pipelines API uses N1 custome machine type
                    start = datetime.strptime(js_get("start", call_metadata), date_format)
                    end = datetime.strptime(js_get("end", call_metadata), date_format)
                    runtime = (end - start).total_seconds()
                    preemptible = bool(int(js_get("runtimeAttributes.preemptible", call_metadata)))
                    disk_description = js_get("runtimeAttributes.disks", call_metadata, default="")
                    if disk_description.startswith("local-disk"):
                        _, size_gb, _ = disk_description.split()
                        disk_size_gb = float(size_gb)
                    else:
                        disk_size_gb = 1.0  # Guess 1GB when disk information is unavailable
                    cost = (costs.GCPCustomN1Cost.estimate(cpus, memory_gb, runtime, preemptible)
                            + costs.PersistentDisk.estimate(disk_size_gb, runtime))
                yield dict(task_name=task_name,
                           cost=cost,
                           number_of_cpus=cpus,
                           memory=memory_gb,
                           disk=disk_size_gb,
                           duration=runtime,
                           call_cached=call_cached)
            except (KeyError, TNUCostException) as exc:
                logger.warning(f"Unable to estimate costs for workflow {workflow_id}: "
                               f"{exc.args[0]}")

def _parse_machine_type(machine_type: str) -> Tuple[int, float]:
    parts = machine_type.split("-", 2)
    if 3 != len(parts) or "custom" != parts[0]:
        raise TNUCostException(f"Cannot estimate costs for machine type '{machine_type}'"
                               "Please contact terra-notebook-utils maintainers to add support")
    try:
        cpus, memory_gb = int(parts[1]), float(parts[2]) / 1024
        return cpus, memory_gb
    except ValueError as exc:
        raise TNUCostException(f"Cannot parse cpus and memory from '{machine_type}'") from exc



import json
import typing
import argparse
from datetime import datetime
from typing import Any, Tuple, List, Dict

from firecloud import fiss

from terra_notebook_utils import workflows
from terra_notebook_utils.cli import Config, dispatch


workflow_cli = dispatch.group("workflows", help=workflows.__doc__)

workspace_args: Dict[str, Dict[str, Any]] = {
    "--workspace": dict(
        type=str,
        default=None,
        help="workspace name. If not provided, the configured CLI workspace will be used"
    ),
    "--workspace-namespace": dict(
        type=str,
        required=False,
        default=Config.info['workspace_namespace'],
        help=("The billing project for GS requests. "
              "If omitted, the CLI configured `workspace_namespace` will be used. "
              "Note that DRS URLs also involve a GS request.")
    )
}

@workflow_cli.command("list-submissions", arguments={**workspace_args})
def list_submissions(args: argparse.Namespace):
    """
    List workflow submissions in chronological order
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    listing = [(s['submissionId'], s['submissionDate'], s['status'])
               for s in workflows.list_submissions(args.workspace, args.workspace_namespace)]
    for submission_id, date, status in sorted(listing, key=lambda item: item[1]):
        print(submission_id, date, status)

@workflow_cli.command("get-submission", arguments={
    "--submission-id": dict(type=str, required=True),
    ** workspace_args
})
def get_submission(args: argparse.Namespace):
    """
    Get information about a submission, including member worklows
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    submission = workflows.get_submission(args.submission_id, args.workspace, args.workspace_namespace)
    print(json.dumps(submission, indent=2))

@workflow_cli.command("get-workflow", arguments={
    "--submission-id": dict(type=str, required=True),
    "--workflow-id": dict(type=str, required=True),
    ** workspace_args
})
def get_workflow(args: argparse.Namespace):
    """
    Get information about a workflow
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    wf = workflows.get_workflow(args.submission_id, args.workflow_id, args.workspace, args.workspace_namespace)
    print(json.dumps(wf, indent=2))

@workflow_cli.command("estimate-submission-cost", arguments={
    "--submission-id": dict(type=str, required=True),
    ** workspace_args
})
def estimate_submission_cost(args: argparse.Namespace):
    """
    Estimate costs for all workflows in a submission
    """
    args.workspace, args.workspace_namespace = Config.resolve(args.workspace, args.workspace_namespace)
    workflows_metadata = workflows.get_all_workflows(args.submission_id, args.workspace, args.workspace_namespace)
    reporter = TXTReport([("workflow_id", 37),
                          ("shard", 6),
                          ("cpus", 5),
                          ("memory (GB)", 12),
                          ("duration (h)", 13),
                          ("call cached", 12),
                          ("cost", 5)])
    reporter.print_headers()
    total = 0
    for workflow_id, workflow_metadata in workflows_metadata.items():
        shard = 1
        for item in workflows.estimate_workflow_cost(workflow_id, workflow_metadata):
            cost, cpus, mem, duration, call_cached = (item[k] for k in ('cost',
                                                                        'number_of_cpus',
                                                                        'memory',
                                                                        'duration',
                                                                        'call_cached'))
            reporter.print_line(workflow_id, shard, cpus, mem, duration / 3600, call_cached, cost)
            total += cost
            shard += 1
    reporter.print_divider()
    reporter.print_line("", "", "", "", "", "", total)

class TXTReport:
    def __init__(self, fields: List[Tuple[str, int]]):
        self.column_headers = [f[0] for f in fields]
        self.widths = [f[1] for f in fields]

    @property
    def width(self) -> int:
        return len(self.line(*[str() for _ in self.widths]))

    def line(self, *vals) -> str:
        assert len(vals) == len(self.widths), f"Expected {len(self.widths)} values"
        parts = list()
        for val, width in zip(vals, self.widths):
            parts.append(self.ff(val, width))
        return " ".join(parts)

    def print_headers(self):
        print(self.line(*self.column_headers))

    def print_line(self, *vals):
        print(self.line(*vals))

    def print_divider(self):
        print("-" * self.width)

    def ff(self, val: Any, width: int, decimals: int=2) -> str:
        if isinstance(val, (str, bool)):
            return f"%{width}s" % val
        elif isinstance(val, int):
            return f"%{width}i" % val
        elif isinstance(val, float):
            return f"%{width}.2f" % round(val, decimals)
        else:
            raise TypeError(f"Unsupported type '{type(val)}'")



"""
Configure the CLI
"""
import os
import json
from typing import Optional, Tuple
from terra_notebook_utils import version

import cli_builder

from terra_notebook_utils import WORKSPACE_NAME, WORKSPACE_GOOGLE_PROJECT


class Config:
    info = dict(workspace=None, workspace_namespace=None)
    path = os.path.join(os.path.expanduser("~"), ".tnu_config")

    @classmethod
    def load(cls):
        if not os.path.isfile(cls.path):
            cls.write()
        with open(cls.path) as fh:
            cls.info = json.loads(fh.read())

    @classmethod
    def write(cls):
        with open(cls.path, "w") as fh:
            fh.write(json.dumps(cls.info, indent=2))

    @classmethod
    def resolve(cls, override_workspace: Optional[str], override_namespace: Optional[str]) -> Tuple[str, str]:
        workspace = override_workspace or (cls.info['workspace'] or WORKSPACE_NAME)
        namespace = override_namespace or (cls.info['workspace_namespace'] or WORKSPACE_GOOGLE_PROJECT)
        if not workspace:
            raise RuntimeError("This command requies a workspace. Either pass in with `--workspace`,"
                               " or configure the CLI (see `tnu config --help`). A default may also be configured by"
                               " setting the `WORKSPACE_NAME` env var")
        if namespace is None:
            from terra_notebook_utils.workspace import get_workspace_namespace
            namespace = get_workspace_namespace(workspace)
        if not namespace:
            raise RuntimeError("This command requies a workspace namespace. Either pass in with `--workspace-namespace`"
                               ", or configure the CLI (see `tnu config --help`). A default may also be"
                               " configued by setting the `WORKSPACE_GOOGLE_PROJECT` env var")
        return workspace, namespace
Config.load()


descr = f"""
    Welcome to the terra-notebook-utils cli, version {version.__version__}
             ___
           _(((,|    What's DNA??
          /  _-\\\\
         / C o\o \\
       _/_    __\ \\     __ __     __ __     __ __     __
      /   \ \___/  )   /--X--\   /--X--\   /--X--\   /--
      |    |\_|\  /   /--/ \--\ /--/ \--\ /--/ \--\ /--/
      |    |#  #|/          \__X__/   \__X__/   \__X__/
      (   /     |
       |  |#  # |
       |  |    #|
       |  | #___n_,_
    ,-/   7-' .     `\\
    `-\...\-_   -  o /
       |#  # `---U--'
       `-v-^-'\/
         \  |_|_ Wny
         (___mnnm
"""
dispatch = cli_builder.Dispatch(description=descr, debug=os.environ.get('TNU_CLI_DEBUG', False))



"""
VCF file utilities
"""
import io
import gzip
from multiprocessing import cpu_count
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import bgzip
import gs_chunked_io as gscio
from gs_chunked_io.async_collections import AsyncQueue

from terra_notebook_utils import gs, drs, IO_CONCURRENCY, WORKSPACE_GOOGLE_PROJECT


cores_available = cpu_count()

class VCFInfo:
    columns = ["chrom", "pos", "id", "ref", "alt", "qual", "filter", "info", "format"]
    chromosomes = [f"chr{i}" for i in range(1, 23)] + ["chrX", "chrY"]

    def __init__(self, fileobj):
        self.header = list()
        for line in fileobj:
            line = line.decode("utf-8").strip()
            if line.startswith("##"):
                self.header.append(line)
            elif line.startswith("#"):
                self.header.append(line)
                self.samples = line.split("\t")[len(self.columns):]
            elif not line:
                continue
            else:
                first_data_line = line
                break
        parts = first_data_line.split("\t", len(self.columns))
        for key, val in zip(self.columns + ["data"], parts):
            setattr(self, key, val)

    @property
    def length(self):
        for line in self.header:
            if "length" in line and f"{self.chrom}," in line:
                length = line.rsplit("=", 1)[-1].replace(">", "")
                return int(length)
        return None

    def print_header(self):
        for line in self.header:
            print(line)

    @classmethod
    def with_bgzip_fileobj(cls, fileobj, read_buf: Optional[memoryview], chunk_size=1024 * 1024):
        if read_buf is None:
            read_buf = memoryview(bytearray(1024 * 1024 * 50))
        with bgzip.BGZipAsyncReaderPreAllocated(fileobj,
                                                read_buf,
                                                num_threads=cores_available,
                                                raw_read_chunk_size=chunk_size) as bgzip_reader:
            with io.BufferedReader(bgzip_reader) as reader:
                return cls(reader)

    @classmethod
    def with_gzip_fileobj(cls, fileobj):
        gzip_reader = gzip.GzipFile(fileobj=fileobj)
        return cls(gzip_reader)

    @classmethod
    def with_blob(cls, blob, read_buf: Optional[memoryview]=None):
        chunk_size = 1024 * 1024
        with ThreadPoolExecutor(max_workers=IO_CONCURRENCY) as e:
            async_queue = AsyncQueue(e, IO_CONCURRENCY)
            try:
                with gscio.Reader(blob, chunk_size, async_queue) as raw:
                    return cls.with_bgzip_fileobj(raw, read_buf, chunk_size)
            except bgzip.BGZIPException:
                with gscio.Reader(blob, chunk_size, async_queue) as raw:
                    return cls.with_gzip_fileobj(raw)

    @classmethod
    def with_file(cls, filepath, read_buf: memoryview=None):
        with open(filepath, "rb") as raw:
            try:
                return cls.with_bgzip_fileobj(raw, read_buf)
            except bgzip.BGZIPException:
                raw.seek(0)
                return cls.with_gzip_fileobj(raw)

def vcf_info(uri: str,
             workspace_namespace: Optional[str]=WORKSPACE_GOOGLE_PROJECT) -> VCFInfo:
    if uri.startswith("drs://"):
        client, drs_info = drs.resolve_drs_for_gs_storage(uri)
        blob = client.bucket(drs_info.bucket_name, user_project=workspace_namespace).get_blob(drs_info.key)
        return VCFInfo.with_blob(blob)
    elif uri.startswith("gs://"):
        bucket, key = uri[5:].split("/", 1)
        blob = gs.get_client().bucket(bucket, user_project=workspace_namespace).get_blob(key)
        return VCFInfo.with_blob(blob)
    elif uri.startswith("s3://"):
        raise ValueError("S3 URIs not supported")
    else:
        return VCFInfo.with_file(uri)

def _headers_equal(a, b):
    for line_a, line_b in zip(a, b):
        if line_a.startswith("##bcftools_viewCommand"):
            # TODO: Include information about which files were combined
            pass
        elif line_a.startswith("##"):
            if line_a != line_b:
                return False
    return True
